// WinampJNI.cpp : Defines the entry point for the DLL application.
//

#include "stdafx.h"
#include <stdlib.h>
#include "wa_ipc.h"

BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
    return TRUE;
}

// Include the header file generated by javah. These are the functions
// we need to define in this file.
#include "WinampHelper.h"

// Include MS headers for using Managed .NET code
#using <mscorlib.dll>
#include <vcclr.h>
#include <string.h>

// Include required for Image operations used in this file
#using <System.Drawing.dll>
#using <System.dll>

using namespace System::Diagnostics;

#include "ManagedToJava.h"

// These struct definitions resolve a TypeLoadException in the managed code that would otherwise
// occur at runtime.
struct _jmethodID {};
struct _jfieldID {};

// Prototypes
void PlayFile(const char *);
void GetWinampInstance(void);
void PromoteSageTV(int mode);
char *JNU_GetStringNativeChars(JNIEnv *, jstring);
void JNU_ThrowByName(JNIEnv *, const char *, const char *);

// Definitions
// WM_COMMAND
#define PREVIOUS_TRACK 40044
#define NEXT_TRACK 40048
#define PLAY 40045
#define PAUSE 40046
#define STOP 40047
#define STOP_WITH_FADE 40147
#define RAISE_VOLUME 40058
#define LOWER_VOLUME 40059
#define STOP_AFTER_CURRENT_TRACK 40157
#define FF_5_SECONDS 40148
#define RW_5_SECONDS 40144
#define START_VISUALIZATION 40192
#define RESTART_PLAYLIST 40154
#define JUMP_TO_END_OF_PLAYLIST 40158
#define CLOSE_WINAMP 40001

// WM_USER
#define CLEAR_PLAYLIST 101
#define CURRENT_STATUS 104
#define CURRENT_TRACK_POSITION 105 	// ms
#define TOTAL_TRACK_DURATION 105 	// seconds
#define SEEK 106
#define SET_VOLUME_ABSOLUTE 122

// Globals
HWND WinampHWND = NULL;
int visstate=0;

void CheckDiagnosticLevel();

#define TIME_ALLOWED_TO_LOAD 8000

DWORD LastStart=0;

int diagnosticlevel = 0;

void CheckDiagnosticLevel()
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering CheckDiagnosticLevel.");

	char *thelevel = getenv("WINAMP_PLUGIN_DIAGNOSTIC_LEVEL");

	diagnosticlevel = 0;

	if (thelevel)
	{
		diagnosticlevel = atoi(thelevel);
	}

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting CheckDiagnosticLevel.");

	return;
}

/*
 * Class:     WinampHelper
 * Method:    WHGetDiagnosticLevel
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_WinampHelper_WHGetDiagnosticLevel
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampHelper_WHGetDiagnosticLevel.");

	CheckDiagnosticLevel();

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampHelper_WHGetDiagnosticLevel.");

	return diagnosticlevel;
}

/*
 * Class:     WinampHelper
 * Method:    WHStartup
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WinampHelper_WHStartup
  (JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampHelper_WHStartup.");

	CheckDiagnosticLevel();

	GetWinampInstance();		

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampHelper_WHStartup.");
	return;
}

/*
 * Class:     WinampHelper
 * Method:    WHShutdown
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WinampHelper_WHShutdown
  (JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampHelper_WHShutdown.");


	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampHelper_WHShutdown.");

	return;
}

void GetWinampInstance()
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering GetWinampInstance.");

	if (diagnosticlevel >= 2)
		Trace::WriteLine("WinampHelper aquiring handle to Winamp.");

	WinampHWND = FindWindow("Winamp v1.x", NULL);

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting GetWinampInstance.");
}

/*
 * Class:     WinampHelper
 * Method:    WHStartVisualization
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WinampHelper_WHStartVisualization
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampHelper_WHStartVisualization.");

	if (diagnosticlevel >= 1)
		Trace::WriteLine("WinampHelper starting visualization.");

	GetWinampInstance();

	if (WinampHWND)
	{
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Sending the start-visualization command.");



		if (diagnosticlevel >= 2)
			Trace::WriteLine("Communication with Winamp is possible; sending start visualization command.");

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_COMMAND,START_VISUALIZATION,0, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the direct-control message; the message has been discarded.");

		// changed:  result = SendMessage(WinampHWND,WM_COMMAND,START_VISUALIZATION,0);

		LastStart = GetTickCount();
		visstate = 1;
	}
	else
	{
		Trace::WriteLine("WinampHelper is not able to communicate with Winamp.");
	}

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampHelper_WHStartVisualization.");
}

/*
 * Class:     WinampHelper
 * Method:    WHStopVisualization()
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WinampHelper_WHStopVisualization
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampHelper_WHStopVisualization.");

	if (diagnosticlevel >= 1)
		Trace::WriteLine("WinampHelper stopping visualization.");

	GetWinampInstance();

	if (WinampHWND)
	{
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Sending the stop-visualization command.");

		if (visstate == 1)
		{


			if (diagnosticlevel >= 2)
				Trace::WriteLine("Communication with Winamp is possible; stopping visualization (reason:  explicit command from SageTV).");

			DWORD MessageResult;
			int alive = SendMessageTimeout(WinampHWND,WM_COMMAND,START_VISUALIZATION,0, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

			if (alive == 0)
				Trace::WriteLine("Winamp did not respond to the direct-control message; the message has been discarded.");

			// changed:  result = SendMessage(WinampHWND,WM_COMMAND,START_VISUALIZATION,0);

			// Having Winamp be the top window is not cool.  Force SageTV to the 
			// foreground.
			PromoteSageTV(1);
			visstate = 0;
		}
	}
	else
	{
		Trace::WriteLine("WinampHelper is not able to communicate with Winamp.");
	}

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampHelper_WHStopVisualization.");
}

void PromoteSageTV(int mode)
{
	HWND HWNDSageTV;
	HWNDSageTV = FindWindow(NULL, "SageTV");
	if (HWNDSageTV)
	{
		if (mode == 1)
		{
			if (diagnosticlevel >= 1)
				Trace::WriteLine("Forcing SageTV to the foreground (cause: visualization explicitly ended by SageTV).");
		}
		
		if (mode == 2)
		{
			if (diagnosticlevel >= 1)
				Trace::WriteLine("Forcing SageTV to the foreground (cause: visualization is not the top window).");
		}
		
		SetForegroundWindow(HWNDSageTV);
	}
	else
	{
		if (mode == 1)
		{
			Trace::WriteLine("An attempt to force SageTV to the foreground was made (because the visualization was explicitly stopped by SageTV) but failed because the SageTV window could not be found.");
		}

		if (mode == 2)
		{
			Trace::WriteLine("An attempt to force SageTV to the foreground was made (because the visualization ceased to be the top window) but failed because the SageTV window could not be found.");
		}
	}
}


/*
 * Class:     WinampHelper
 * Method:    WHCheckVisualization
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_WinampHelper_WHCheckVisualization
  (JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampHelper_WHCheckVisualization.");

	// Is it time to check for the visualization exiting?
	if (GetTickCount() - LastStart > TIME_ALLOWED_TO_LOAD)
	{

		// Figure out what window is the current foreground window.
		HWND CurrentFGHWND;
		char this_title[2048];

		CurrentFGHWND = GetForegroundWindow();
		GetWindowText(CurrentFGHWND, this_title, sizeof(this_title));

		// Build a message to display.
		char TheMessage[512];
		strcpy(TheMessage, "The WinampHelper sees the current foreground window title as '");
		strcat(TheMessage, this_title);
		strcat(TheMessage, "'.");

		if (diagnosticlevel >= 2)
		{
			Trace::WriteLine("Checking if the visualization has ended.");
			Trace::WriteLine(TheMessage);
		}

		if (strstr(this_title, "Winamp") != NULL)
		{
			// The Winamp window has somehow become the foreground window.  We don't want that.
			if (diagnosticlevel >= 1)
				Trace::WriteLine("The current foreground window contains the text 'Winamp' (ie: it is not the visualization).  Activating SageTV.");
			PromoteSageTV(2);
			return 1;
		}
		else
		{
			if (diagnosticlevel >= 2)
				Trace::WriteLine("The current foreground window does not contain the text 'Winamp' -- no action is required.");
		}
	}
	else
	{
		if (diagnosticlevel >= 2)
			Trace::WriteLine("The visualization is still in its loading grace period, so automatic visualization state checking is disabled -- no action is required.");
	}

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampHelper_WHCheckVisualization.");
	return 0;
}

/*
 * Class:     WinampHelper
 * Method:    WHDuration
 * Signature: ()J
 */
JNIEXPORT jlong JNICALL Java_WinampHelper_WHDuration
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampHelperin_WHDuration.");

	GetWinampInstance();

	if (WinampHWND)
	{
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Sending the get-duration command.");

		LRESULT result;
		LRESULT result0;

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_USER,1,TOTAL_TRACK_DURATION, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		result = MessageResult;

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the direct-control message; the message has been discarded.");

		// changed:  result = SendMessage(WinampHWND,WM_USER,1,TOTAL_TRACK_DURATION);

		DWORD MessageResult2;
		int alive2 = SendMessageTimeout(WinampHWND,WM_COMMAND,START_VISUALIZATION,0, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult2);

		if (alive2 == 0)
			Trace::WriteLine("Winamp did not respond to the direct-control message; the message has been discarded.");

		result0 = MessageResult2;

		// changed:  result0 = SendMessage(WinampHWND,WM_USER,0,CURRENT_TRACK_POSITION);

		if (diagnosticlevel >= 2)
			Trace::WriteLine("Exiting Java_WinampHelper_WHDuration.");

		// Never return a negative value.
		if (result >= 0)
			return (result * 1000);
		else
			return 0;

	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampHelper_WHDuration.");
	return 0;
}

/*
 * Class:     WinampHelper
 * Method:    WHElapsed
 * Signature: ()J
 */
JNIEXPORT jlong JNICALL Java_WinampHelper_WHElapsed
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampHelper_WHElapsed.");

	GetWinampInstance();

	if (WinampHWND)
	{
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Sending the get-elapsed command.");

		LRESULT result;
		LRESULT result2;

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_USER,0,CURRENT_TRACK_POSITION, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the direct-control message; the message has been discarded.");

		result = MessageResult;

		// changed:  result = SendMessage(WinampHWND,WM_USER,0,CURRENT_TRACK_POSITION);

		DWORD MessageResult2;
		int alive2 = SendMessageTimeout(WinampHWND,WM_USER,1,TOTAL_TRACK_DURATION, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult2);

		if (alive2 == 0)
			Trace::WriteLine("Winamp did not respond to the direct-control message; the message has been discarded.");

		result2 = MessageResult2;

		// changed:  result2 = SendMessage(WinampHWND,WM_USER,1,TOTAL_TRACK_DURATION);

		result2 *= 1000;

		if (diagnosticlevel >= 2)
			Trace::WriteLine("Exiting Java_WinampHelper_WHElapsed.");

		// Never return a negative value.
		if (result >= 0)
			return result;
		else
			return 0;
	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampHelper_WHElapsed.");
	return 0;
}

/*
 * Class:     WinampHelper
 * Method:    WHDirectControlPlayFileOrURL
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_WinampHelper_WHDirectControlPlayFile
  (JNIEnv *env, jobject jo, jstring TheFile)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampHelper_WHDirectControlPlayFileOrURL.");

	if (diagnosticlevel >= 1)
	{
		Trace::WriteLine("Direct Control Command Issued -- Playing:  ");
		Trace::WriteLine(JStringToMgdString(env, TheFile));
	}

	// const char *str = env->GetStringUTFChars(TheFile, 0);
	const char *str = JNU_GetStringNativeChars(env, TheFile);

	PlayFile(str);

	env->ReleaseStringUTFChars(TheFile, str);

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampHelper_WHDirectControlPlayFileOrURL.");
	return;
}

void PlayFile(const char *filename)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering PlayFile.");

	GetWinampInstance();

	if (WinampHWND)
	{
		COPYDATASTRUCT cds;
	
		cds.dwData = IPC_PLAYFILE;
		cds.lpData = (char *)filename;
		cds.cbData = (unsigned int)strlen((char*)cds.lpData) + 1; // Include space for null char at the end

		// Clear the playlist.
		if (diagnosticlevel >= 1)
			Trace::WriteLine("Direct Control Command Issued (implied) -- Clear Playlist.");

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_USER,0,CLEAR_PLAYLIST, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the direct-control message; the message has been discarded.");

		// changed:  SendMessage(WinampHWND,WM_USER,0,CLEAR_PLAYLIST);

		// Load the track.
		if (diagnosticlevel >= 1)
			Trace::WriteLine("Direct Control Command Issued -- Enqueue file.");

		DWORD MessageResult2;
		int alive2 = SendMessageTimeout(WinampHWND,WM_COPYDATA,WPARAM(NULL),(LPARAM)&cds, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult2);

		if (alive2 == 0)
			Trace::WriteLine("Winamp did not respond to the direct-control message; the message has been discarded.");

		// changed:  SendMessage(WinampHWND, WM_COPYDATA, WPARAM(NULL), (LPARAM)&cds);

		DWORD MessageResult3;
		int alive3 = SendMessageTimeout(WinampHWND,WM_COMMAND,PLAY,0, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult3);

		if (alive3 == 0)
			Trace::WriteLine("Winamp did not respond to the direct-control message; the message has been discarded.");
		
		// changed:  result = SendMessage(WinampHWND,WM_COMMAND,PLAY,0);
	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting PlayFile.");
}



/*
 * Class:     WinampHelper
 * Method:    WHDirectControlPlay
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WinampHelper_WHDirectControlPlay
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampHelper_WHDirectControlPlay.");

	GetWinampInstance();

	if (WinampHWND)
	{
		if (diagnosticlevel >= 1)
			Trace::WriteLine("Direct Control Command Issued -- Play.");

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_COMMAND, PLAY,0, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the direct-control message; the message has been discarded.");

		// changed:  result = SendMessage(WinampHWND,WM_COMMAND,PLAY,0);
	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampHelper_WHDirectControlPlay.");
}



/*
 * Class:     WinampHelper
 * Method:    WHDirectControlPause
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WinampHelper_WHDirectControlPause
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampHelper_WHDirectControlPause.");

	GetWinampInstance();

	if (WinampHWND)
	{
		if (diagnosticlevel >= 1)
			Trace::WriteLine("Direct Control Command Issued -- Pause.");

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_COMMAND,PAUSE,0, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the direct-control message; the message has been discarded.");

		// changed:  result = SendMessage(WinampHWND,WM_COMMAND,PAUSE,0);
	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampHelper_WHDirectControlPause.");
}



/*
 * Class:     WinampHelper
 * Method:    WHDirectControlStop
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WinampHelper_WHDirectControlStop
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampHelper_WHDirectControlStop.");

	GetWinampInstance();

	if (WinampHWND)
	{
		if (diagnosticlevel >= 1)
			Trace::WriteLine("Direct Control Command Issued -- Stop.");

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_COMMAND,STOP,0, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the direct-control message; the message has been discarded.");

		// changed:  result = SendMessage(WinampHWND,WM_COMMAND,STOP,0);
	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampHelper_WHDirectControlStop.");
}



/*
 * Class:     WinampHelper
 * Method:    WHDirectControlVolumeUp
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WinampHelper_WHDirectControlVolumeUp
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampHelper_WHDirectControlVolumeUp.");

	GetWinampInstance();

	if (WinampHWND)
	{
		if (diagnosticlevel >= 1)
			Trace::WriteLine("Direct Control Command Issued -- Volume Up 1 Percent");

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_COMMAND,RAISE_VOLUME,0, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the direct-control message; the message has been discarded.");

		// changed:  result = SendMessage(WinampHWND,WM_COMMAND,RAISE_VOLUME,0);
	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampHelper_WHDirectControlVolumeUp.");
}




/*
 * Class:     WinampHelper
 * Method:    WHDirectControlVolumeDown
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WinampHelper_WHDirectControlVolumeDown
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampHelper_WHDirectControlVolumeDown.");

	GetWinampInstance();

	if (WinampHWND)
	{
		if (diagnosticlevel >= 1)
			Trace::WriteLine("Direct Control Command Issued -- Volume Down 1 Percent");

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_COMMAND,LOWER_VOLUME,0, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the direct-control message; the message has been discarded.");

		// changed:  result = SendMessage(WinampHWND,WM_COMMAND,LOWER_VOLUME,0);
	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampHelper_WHDirectControlVolumeDown.");
}



/*
 * Class:     WinampHelper
 * Method:    WHDirectControlSetVolume
 * Signature: (F)F
 */
JNIEXPORT jfloat JNICALL Java_WinampHelper_WHDirectControlSetVolume
  (JNIEnv *env, jobject jo, jfloat TheVolume)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampHelper_WHDirectControlSetVolume.");

	GetWinampInstance();

	if (WinampHWND)
	{
		int volume=255;

		volume = (float)TheVolume * (float)255;

		if (diagnosticlevel >= 1)
			Trace::WriteLine("Direct Control Command Issued -- Set Volume Absolute");

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_USER,volume,SET_VOLUME_ABSOLUTE, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the direct-control message; the message has been discarded.");

		// changed:  result = SendMessage(WinampHWND,WM_USER,volume,SET_VOLUME_ABSOLUTE);
	}

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampHelper_WHDirectControlSetVolume.");
	return TheVolume;
}

/*
 * Class:     WinampHelper
 * Method:    WHDirectControlSeek
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_WinampHelper_WHDirectControlSeek
  (JNIEnv *env, jobject jo, jlong SeekTime)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampHelper_WHDirectControlSeek.");

	GetWinampInstance();

	if (WinampHWND)
	{
		if (diagnosticlevel >= 1)
			Trace::WriteLine("Direct Control Command Issued -- Seek");

		// Do not allow seeking to a negative position.
		if (SeekTime < 0)
			SeekTime = 0;

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_USER,SeekTime,SEEK, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the direct-control message; the message has been discarded.");

		// changed:  result = SendMessage(WinampHWND,WM_USER,SeekTime,SEEK);

		return SeekTime;
	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampHelper_WHDirectControlSeek.");
	return SeekTime;
}



void JNU_ThrowByName(JNIEnv *env, const char *name, const char *msg)
{
	jclass cls;

	cls = env->FindClass(name);

	/* if cls is NULL, an exception has already been thrown */

	if (cls != NULL) {
		env->ThrowNew(cls, msg);
	}

	/* free the local ref */
	env->DeleteLocalRef(cls);
}

char *JNU_GetStringNativeChars(JNIEnv *env, jstring jstr)
{
	jbyteArray bytes = 0;
	jthrowable exc;
	char *result = 0;
	jclass jcls_str;
	jmethodID MID_String_getBytes;

	if (env->EnsureLocalCapacity(2) < 0) {
		return 0; /* out of memory error */
	}

	jcls_str = env->FindClass("java/lang/String");

	MID_String_getBytes = env->GetMethodID(jcls_str, "getBytes", "()[B");

	bytes = (jbyteArray)env->CallObjectMethod(jstr, MID_String_getBytes);
	exc = env->ExceptionOccurred();

	if (!exc) {
		jint len = env->GetArrayLength(bytes);
		result = (char *)malloc(len + 1);

		if (result == 0) {
			JNU_ThrowByName(env, "java/lang/OutOfMemoryError", 0);
			env->DeleteLocalRef(bytes);
			return 0;
		}

		env->GetByteArrayRegion(bytes, 0, len, (jbyte *)result);

		result[len] = 0; /* NULL-terminate */
	} else {
		env->DeleteLocalRef(exc);
	}

	env->DeleteLocalRef(bytes);
	return (char*)result;
}

/*
 * Class:     WinampHelper
 * Method:    WHDirectControlStopWithFade
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WinampHelper_WHDirectControlStopWithFade
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampHelper_WHDirectControlStopWithFade.");

	GetWinampInstance();

	if (WinampHWND)
	{
		if (diagnosticlevel >= 1)
			Trace::WriteLine("Direct Control Command Issued -- Stop (Fade).");

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_COMMAND,STOP_WITH_FADE,0, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the direct-control message; the message has been discarded.");

		// changed:  result = SendMessage(WinampHWND,WM_COMMAND,STOP_WITH_FADE,0);
	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampHelper_WHDirectControlStopWithFade.");
}



/*
 * Class:     WinampHelper
 * Method:    WHDirectControlStopAfterCurrentTrack
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WinampHelper_WHDirectControlStopAfterCurrentTrack
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampHelper_WHDirectControlStopAfterCurrentTrack.");

	GetWinampInstance();

	if (WinampHWND)
	{
		if (diagnosticlevel >= 1)
			Trace::WriteLine("Direct Control Command Issued -- Stop After Current Track.");

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_COMMAND,STOP_AFTER_CURRENT_TRACK,0, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the direct-control message; the message has been discarded.");

		// changed:  result = SendMessage(WinampHWND,WM_COMMAND,STOP_AFTER_CURRENT_TRACK,0);
	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampHelper_WHDirectControlStopAfterCurrentTrack.");
}



/*
 * Class:     WinampHelper
 * Method:    WHDirectControlFF5
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WinampHelper_WHDirectControlFF5
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampHelper_WHDirectControlFF5.");

	GetWinampInstance();

	if (WinampHWND)
	{
		if (diagnosticlevel >= 1)
			Trace::WriteLine("Direct Control Command Issued -- Fast Forward 5 Seconds.");

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_COMMAND,FF_5_SECONDS,0, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the direct-control message; the message has been discarded.");

		// changed:  result = SendMessage(WinampHWND,WM_COMMAND,FF_5_SECONDS,0);
	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampHelper_WHDirectControlFF5.");
}



/*
 * Class:     WinampHelper
 * Method:    WHDirectControlRW5
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WinampHelper_WHDirectControlRW5
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampHelper_WHDirectControlRW5.");

	GetWinampInstance();

	if (WinampHWND)
	{
		if (diagnosticlevel >= 1)
			Trace::WriteLine("Direct Control Command Issued -- Rewind 5 Seconds.");

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_COMMAND,RW_5_SECONDS,0, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the direct-control message; the message has been discarded.");

		// changed:  result = SendMessage(WinampHWND,WM_COMMAND,RW_5_SECONDS,0);
	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampHelper_WHDirectControlRW5.");
}




/*
 * Class:     WinampHelper
 * Method:    WHDirectControlGetPlaystate
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_WinampHelper_WHDirectControlGetPlaystate
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampHelper_WHDirectControlGetPlaystate.");

	GetWinampInstance();

	if (WinampHWND)
	{
		if (diagnosticlevel >= 1)
			Trace::WriteLine("Direct Control Command Issued -- Current Status.");

		LRESULT result;

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_USER,0,CURRENT_STATUS, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the direct-control message; the message has been discarded.");

		// changed:  result = SendMessage(WinampHWND,WM_USER,0,CURRENT_STATUS);

		return (jint)result;
	}

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampHelper_WHDirectControlGetPlaystate.");

	return 0;
}


