// WinampJNI.cpp : Defines the entry point for the DLL application.
//

#include "stdafx.h"
#include <stdlib.h>
#include "wa_ipc.h"

BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
    return TRUE;
}

// Include the header file generated by javah. These are the functions
// we need to define in this file.
#include "WinampPlugin.h"

// Include MS headers for using Managed .NET code
#using <mscorlib.dll>
#include <vcclr.h>
#include <string.h>

// Include required for Image operations used in this file
#using <System.Drawing.dll>
#using <System.dll>

using namespace System::Diagnostics;

#include "ManagedToJava.h"

// These struct definitions resolve a TypeLoadException in the managed code that would otherwise
// occur at runtime.
struct _jmethodID {};
struct _jfieldID {};

// Prototypes
void PlayFile(const char *);
void GetWinampInstance(void);
void SendNudge();
void CheckDiagnosticLevel();

/*

// Uncomment the following code to support international characters.

char *JNU_GetStringNativeChars(JNIEnv *, jstring);
void JNU_ThrowByName(JNIEnv *, const char *, const char *);

*/

// Constants (WM_COMMAND)
#define PREVIOUS_TRACK 40044
#define NEXT_TRACK 40048
#define PLAY 40045
#define PAUSE 40046
#define STOP 40047
#define STOP_WITH_FADE 40147
#define RAISE_VOLUME 40058
#define LOWER_VOLUME 40059
#define STOP_AFTER_CURRENT_TRACK 40157
#define FF_5_SECONDS 40148
#define RW_5_SECONDS 40144
#define START_VISUALIZATION 40192
#define RESTART_PLAYLIST 40154
#define JUMP_TO_END_OF_PLAYLIST 40158
#define CLOSE_WINAMP 40001

// Constants (WM_USER)
#define CLEAR_PLAYLIST 101
#define CURRENT_STATUS 104
#define CURRENT_TRACK_POSITION 105 	// ms
#define TOTAL_TRACK_DURATION 105 	// seconds
#define SEEK 106
#define SET_VOLUME_ABSOLUTE 122

// Constants (Timing)
#define TIME_ALLOWED_TO_LOAD 5000
#define TIME_ALLOWED_FOR_WINAMP_TO_LOAD 15000
#define TIME_BETWEEN_NUDGES 5000

// Constants (States)
int NO_STATE = 0;
int LOADED_STATE = 1;
int PLAY_STATE = 2;
int PAUSE_STATE = 3;
int STOPPED_STATE = 4;
int EOS_STATE = 5;

// Globals (Conrtols)
int state=0;
int muted = 0;
int volume=255;
int diagnosticlevel = 0;

// Globals (Tasks)
HWND WinampHWND = NULL;
STARTUPINFO si;
PROCESS_INFORMATION pi;

// Globals (Timing)
DWORD LastLoad=0;
DWORD LastNudge=0;
DWORD LastWinampStart=0;

// Globals (Feeble Fake Mutex)
int busy = 0;

/*
 * Class:     WinampPlugin
 * Method:    WPStartup
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WinampPlugin_WPStartup
  (JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampPlugin_WPStartup.");

	CheckDiagnosticLevel();

	GetWinampInstance();		

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampPlugin_WPStartup.");
	return;
}

/*
 * Class:     WinampPlugin
 * Method:    WPShutdown
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WinampPlugin_WPShutdown
  (JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampPlugin_WPShutdown.");

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampPlugin_WPShutdown.");
	return;
}

/*
 * Class:     WinampPlugin
 * Method:    WPPlayFile
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_WinampPlugin_WPPlayFile
  (JNIEnv *env, jobject jo, jstring TheFile)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampPlugin_WPPlayFile.");

	LastLoad = GetTickCount();

	if (state == PLAY_STATE)
	{
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Warning:  Instructed to load a file while another file is actively playing.");
	}

	if (diagnosticlevel >= 1)
		Trace::WriteLine("Playing:  ");
	if (diagnosticlevel >= 1)
		Trace::WriteLine(JStringToMgdString(env, TheFile));

	const char *str = env->GetStringUTFChars(TheFile, 0);

	/*

	// Uncomment the following code to support international characters.

	const char *str = JNU_GetStringNativeChars(env, TheFile);
	*/

	PlayFile(str);

	env->ReleaseStringUTFChars(TheFile, str);

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampPlugin_WPPlayFile.");
	return;
}

void PlayFile(const char *filename)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering WPPlayFile.");

	GetWinampInstance();

	if (WinampHWND)
	{
		COPYDATASTRUCT cds;
	
		cds.dwData = IPC_PLAYFILE;
		cds.lpData = (char *)filename;
		cds.cbData = (unsigned int)strlen((char*)cds.lpData) + 1; // Include space for null char at the end

		if (diagnosticlevel >= 2)
			Trace::WriteLine("Waiting for the Winamp playlist to be ready.");

		// Clear the playlist.
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Sending CLEAR_PLAYLIST message to Winamp and waiting for a response.");

		GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
				     // a message.  This is necessary because Winamp could crash between the
				     // last call to GetWinampInstance() and this message send.

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_USER,0,CLEAR_PLAYLIST, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the CLEAR_PLAYLIST message; the message has been discarded.  (1)");

		// changed:  SendMessage(WinampHWND,WM_USER,0,CLEAR_PLAYLIST);

		if (diagnosticlevel >= 2)
			Trace::WriteLine("The response for the CLEAR_PLAYLIST message has been received.");

		// Load the track.
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Sending WM_COPYDATA message to Winamp and waiting for a response.");

		GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
				     // a message.  This is necessary because Winamp could crash between the
				     // last call to GetWinampInstance() and this message send.

		DWORD MessageResult2;
		int alive2 = SendMessageTimeout(WinampHWND,WM_COPYDATA,WPARAM(NULL),(LPARAM)&cds, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult2);

		if (alive2 == 0)
			Trace::WriteLine("Winamp did not respond to the WM_COPYDATA message; the message has been discarded.  (2)");

		// changed:  SendMessage(WinampHWND, WM_COPYDATA, WPARAM(NULL), (LPARAM)&cds);

		if (diagnosticlevel >= 2)
			Trace::WriteLine("The response for the WM_COPYDATA message has been received.");

		LRESULT result;
		
		// Set the volume to nothing.
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Sending SET_VOLUME_ABSOLUTE message to Winamp and waiting for a response.");

		GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
				     // a message.  This is necessary because Winamp could crash between the
				     // last call to GetWinampInstance() and this message send.

		DWORD MessageResult3;
		int alive3 = SendMessageTimeout(WinampHWND,WM_USER,0,SET_VOLUME_ABSOLUTE, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult3);

		if (alive3 == 0)
			Trace::WriteLine("Winamp did not respond to the SET_VOLUME_ABSOLUTE message; the message has been discarded.  (3)");

		result = MessageResult3;

		// changed:  result = SendMessage(WinampHWND,WM_USER,0,SET_VOLUME_ABSOLUTE);

		if (diagnosticlevel >= 2)
			Trace::WriteLine("The response for the SET_VOLUME_ABSOLUTE message has been received.");

		// Start playing the track and then immediately pause.  This should make the duration available.
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Sending PLAY message to Winamp and waiting for a response.");

		GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
				     // a message.  This is necessary because Winamp could crash between the
				     // last call to GetWinampInstance() and this message send.

		DWORD MessageResult4;
		int alive4 = SendMessageTimeout(WinampHWND,WM_COMMAND,PLAY,0, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult4);

		if (alive4 == 0)
			Trace::WriteLine("Winamp did not respond to the PLAY message; the message has been discarded.  (4)");

		result = MessageResult4;

		// changed:  result = SendMessage(WinampHWND,WM_COMMAND,PLAY,0);

		if (diagnosticlevel >= 2)
			Trace::WriteLine("The response for the PLAY message has been received.");

		if (diagnosticlevel >= 2)
			Trace::WriteLine("Sending STOP_AFTER_CURRENT_TRACK message to Winamp and waiting for a response.");

		GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
				     // a message.  This is necessary because Winamp could crash between the
				     // last call to GetWinampInstance() and this message send.

		DWORD MessageResult5;
		int alive5 = SendMessageTimeout(WinampHWND,WM_COMMAND,STOP_AFTER_CURRENT_TRACK,0, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult5);

		if (alive5 == 0)
			Trace::WriteLine("Winamp did not respond to the STOP_AFTER_CURRENT_TRACK message; the message has been discarded.  (5)");

		result = MessageResult5;

		// changed:  result = SendMessage(WinampHWND,WM_COMMAND,STOP_AFTER_CURRENT_TRACK,0);

		if (diagnosticlevel >= 2)
			Trace::WriteLine("The response for the STOP_AFTER_CURRENT_TRACK message has been received.");

		// Get the duration of the track in s.
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Sending TOTAL_TRACK_DURATION message to Winamp and waiting for a response.");

		GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
				     // a message.  This is necessary because Winamp could crash between the
				     // last call to GetWinampInstance() and this message send.

		DWORD MessageResult6;
		int alive6 = SendMessageTimeout(WinampHWND,WM_USER,1,TOTAL_TRACK_DURATION, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult6);

		if (alive6 == 0)
			Trace::WriteLine("Winamp did not respond to the TOTAL_TRACK_DURATION message; the message has been discarded.  (6)");

		result = MessageResult6;

		// changed:  result = SendMessage(WinampHWND,WM_USER,1,TOTAL_TRACK_DURATION);

		if (diagnosticlevel >= 2)
			Trace::WriteLine("The response for the TOTAL_TRACK_DURATION message has been received.");

		DWORD LoadingStart = GetTickCount();

		while (GetTickCount() - LoadingStart <= TIME_ALLOWED_TO_LOAD)
		{
			// Get the duration of the track in s.
			if (diagnosticlevel >= 2)
				Trace::WriteLine("Sending TOTAL_TRACK_DURATION message to Winamp and waiting for a response.");

			GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
					     // a message.  This is necessary because Winamp could crash between the
					     // last call to GetWinampInstance() and this message send.

			DWORD MessageResult7;
			int alive7 = SendMessageTimeout(WinampHWND,WM_USER,1,TOTAL_TRACK_DURATION, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult7);

			if (alive7 == 0)
				Trace::WriteLine("Winamp did not respond to the TOTAL_TRACK_DURATION message; the message has been discarded.  (7)");
	
			result = MessageResult7;

			// changed:  result = SendMessage(WinampHWND,WM_USER,1,TOTAL_TRACK_DURATION);

			if (diagnosticlevel >= 2)
				Trace::WriteLine("The response for the TOTAL_TRACK_DURATION message has been received.");

			if (result >= 0)
			{
				if (diagnosticlevel >= 2)
					Trace::WriteLine("Waiting for the Winamp playlist to be ready.");
				break;
			}

		}

		if (result >= 0)
		{
			if (diagnosticlevel >= 2)
				Trace::WriteLine("The playlist is ready - got a duration back from Winamp that is >= 0.");
		}
		else
		{
			if (diagnosticlevel >= 2)
				Trace::WriteLine("The duration Winamp returned during loading is < 0.");
		}

		// Restore the volume.
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Sending SET_VOLUME_ABSOLUTE message to Winamp and waiting for a response.");

		GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
				     // a message.  This is necessary because Winamp could crash between the
				     // last call to GetWinampInstance() and this message send.

		DWORD MessageResult8;
		int alive8 = SendMessageTimeout(WinampHWND,WM_USER,volume,SET_VOLUME_ABSOLUTE, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult8);

		if (alive8 == 0)
			Trace::WriteLine("Winamp did not respond to the SET_VOLUME_ABSOLUTE message; the message has been discarded.  (8)");

		result = MessageResult8;

		// changed:  result = SendMessage(WinampHWND,WM_USER,volume,SET_VOLUME_ABSOLUTE);

		if (diagnosticlevel >= 2)
			Trace::WriteLine("The response for the SET_VOLUME_ABSOLUTE message has been received.");

		state = LOADED_STATE;

		if (diagnosticlevel >= 2)
			Trace::WriteLine("File loaded.");

		if (diagnosticlevel >= 1)
			Trace::WriteLine("MediaPlayer state changed in WPPlayFile() to LOADED_STATE.");
	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting WPPlayFile.");
	return;
}

void StartWinamp()
{
	
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering StartWinamp().");

	if (GetTickCount() - LastWinampStart > TIME_ALLOWED_FOR_WINAMP_TO_LOAD)
	{
		LastWinampStart = GetTickCount();

		if (diagnosticlevel >= 1)
			Trace::WriteLine("The WinampPlugin is starting an instance of Winamp.");

		ZeroMemory( &si, sizeof(si) );
		si.cb = sizeof(si);
		ZeroMemory( &pi, sizeof(pi) );

		si.dwFlags = STARTF_USESHOWWINDOW;

		si.wShowWindow = SW_SHOWMINNOACTIVE; // or SW_HIDE

		// Start the child process. 
		if (!CreateProcess( NULL,	  		 	// No module name (use command line). 
        		TEXT("C:\\Program Files\\Winamp\\winamp.exe"), 	// Command line. 
		        NULL,             				// Process handle not inheritable. 
	        	NULL,             				// Thread handle not inheritable. 
	        	FALSE,            				// Set handle inheritance to FALSE. 
        		0,                				// No creation flags. 
		        NULL,             				// Use parent's environment block. 
        		NULL,           	  			// Use parent's starting directory. 
		        &si,           		   			// Pointer to STARTUPINFO structure.
	        	&pi))             				// Pointer to PROCESS_INFORMATION structure.
		{
			Trace::WriteLine("The WinampPlugin attempted to launch Winamp, but failed to do so.");
       		 	return;
		}
		else
		{
			if (diagnosticlevel >= 1)
				Trace::WriteLine("The WinampPlugin successfully started an instance of Winamp; waiting for it to finish loading.");

			while (!WinampHWND)
			{
				Sleep(100);
				WinampHWND = FindWindow("Winamp v1.x", NULL);
			}

			Sleep(3000);

			if (diagnosticlevel >= 1)
				Trace::WriteLine("The new instance of Winamp has created its window.  Sending it a message to determine if it is ready to receive commands.");

			// We want to chill until we get a real response from Winamp.
			LRESULT result;

			// desired choke point
			DWORD MessageResult;
			int alive = SendMessageTimeout(WinampHWND,WM_WA_IPC,0,IPC_GETVERSION, SMTO_ABORTIFHUNG | SMTO_NORMAL, 5000, &MessageResult);

			if (alive == 0)
				Trace::WriteLine("Winamp did not respond to the IPC_GETVERSION; the message has been discarded.  (0)");

			result = MessageResult;

			// changed:  result = SendMessage(WinampHWND,WM_WA_IPC,0,IPC_GETVERSION);

			if (diagnosticlevel >= 1)
				Trace::WriteLine("The new Winamp instance is now ready for use.");
		}

	}
	else
	{
		if (diagnosticlevel >= 1)
			Trace::WriteLine("A request to load Winamp was made, but is being ignore because it was made within the Winamp startup grace period.");
	}

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting StartWinamp.");
	
	return;
}

void TerminateWinamp()
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering TerminateWinamp.");

	if (GetTickCount() - LastWinampStart > TIME_ALLOWED_FOR_WINAMP_TO_LOAD)
	{
		if (diagnosticlevel >= 1)
			Trace::WriteLine("The Winamp instance is being terminated.");

		// Kill Winamp with EXTREME prejudice.  
		TerminateProcess(pi.hProcess, 0);

		// Close process and thread handles. 
		CloseHandle(pi.hProcess);
		CloseHandle(pi.hThread);
	}
	else
	{
		Trace::WriteLine("A request to terminate Winamp is being ignored because it is made within the Winamp startup grace period.");
	}

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting TerminateWinamp().");

	return;
}

void GetWinampInstance()
{

	// block aggressively to force synchronicity
	while (busy>0)
	{
		if (diagnosticlevel >= 3)
			Trace::WriteLine("Waiting for Winamp mutex to clear.");
	}

	busy++;

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering GetWinampInstance.");

	WinampHWND = FindWindow("Winamp v1.x", NULL);

	if (!WinampHWND)
	{
		// Winamp is not running.
		Trace::WriteLine("Winamp does not appear to be running!  A new instance of Winamp will be started.");

		// So start it!
		StartWinamp();

		WinampHWND = FindWindow("Winamp v1.x", NULL);
	}
	else
	{
		// Send a test message to see if Winamp is alive.
		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_WA_IPC,0,IPC_GETVERSION, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
		{
			Trace::WriteLine("Winamp appears to be running, but unresponsive!  The existing instance of Winamp will be terminated and a new instance launched.");

			// Do not use the existing Winamp handle.
			WinampHWND = NULL;

			TerminateWinamp();
		}
		else
		{
			if (diagnosticlevel >= 2)
				Trace::WriteLine("Winamp appears to be running and responsive.");
		}
	}

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting GetWinampInstance.");
	busy--;

	return;
}

/*
 * Class:     WinampPlugin
 * Method:    WPPlay
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WinampPlugin_WPPlay
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampPlugin_WPPlay.");

	GetWinampInstance();

	if (state == PLAY_STATE)
	{
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Instructed to start playback while playback is already in progress; ignoring.");
	}
	else
	{
		if (WinampHWND)
		{
			state = PLAY_STATE;

			LRESULT result;
			if (diagnosticlevel >= 2)
				Trace::WriteLine("Sending PLAY message to Winamp and waiting for a response.");

			GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
					     // a message.  This is necessary because Winamp could crash between the
					     // last call to GetWinampInstance() and this message send.

			DWORD MessageResult;
			int alive = SendMessageTimeout(WinampHWND,WM_COMMAND,PLAY,0, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

			if (alive == 0)
				Trace::WriteLine("Winamp did not respond to the PLAY; the message has been discarded.  (9)");

			result = MessageResult;

			// changed:  result = SendMessage(WinampHWND,WM_COMMAND,PLAY,0);

			if (diagnosticlevel >= 2)
				Trace::WriteLine("The response for the PLAY message has been recevied.");

			if (diagnosticlevel >= 2)
				Trace::WriteLine("Sending STOP_AFTER_CURRENT_TRACK message to Winamp and waiting for a response.");

			GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
					     // a message.  This is necessary because Winamp could crash between the
					     // last call to GetWinampInstance() and this message send.

			DWORD MessageResult2;
			int alive2 = SendMessageTimeout(WinampHWND,WM_COMMAND,STOP_AFTER_CURRENT_TRACK,0, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult2);
	
			if (alive2 == 0)
				Trace::WriteLine("Winamp did not respond to the STOP_AFTER_CURRENT_TRACK message; the message has been discarded.  (10)");

			result = MessageResult2;

			// changed:  result = SendMessage(WinampHWND,WM_COMMAND,STOP_AFTER_CURRENT_TRACK,0);

			if (diagnosticlevel >= 2)
				Trace::WriteLine("The response for the STOP_AFTER_CURRENT_TRACK message has been received.");

			if (diagnosticlevel >= 1)
				Trace::WriteLine("MediaPlayer state changed in WPPlay() to PLAY_STATE.");
		}
	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampPlugin_WPPlay.");
	return;
}

/*
 * Class:     WinampPlugin
 * Method:    WPPause
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WinampPlugin_WPPause
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampPlugin_WPPause.");

	GetWinampInstance();

	if (WinampHWND)
	{
		state = PAUSE_STATE;

		LRESULT result;
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Sending PAUSE message to Winamp and waiting for a response.");

		GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
				     // a message.  This is necessary because Winamp could crash between the
				     // last call to GetWinampInstance() and this message send.

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_COMMAND,PAUSE,0, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the PAUSE message; the message has been discarded.  (11)");

		result = MessageResult;

		// changed:  result = SendMessage(WinampHWND,WM_COMMAND,PAUSE,0);

		if (diagnosticlevel >= 2)
			Trace::WriteLine("The response for the PAUSE message has been received.");

		if (diagnosticlevel >= 1)
			Trace::WriteLine("MediaPlayer state changed in WPPause() to PAUSE_STATE.");
	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampPlugin_WPPause.");
	return;
}

/*
 * Class:     WinampPlugin
 * Method:    WPStop
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WinampPlugin_WPStop
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampPlugin_WPStop.");

	GetWinampInstance();

	if (state == STOPPED_STATE)
	{
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Instructed to stop playback while playback is already stopped; ignoring.");
	}
	else
	{
		if (WinampHWND)
		{
			state = STOPPED_STATE;

			LRESULT result;
			if (diagnosticlevel >= 2)
				Trace::WriteLine("Sending STOP message to Winamp and waiting for a response.");

			GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
					     // a message.  This is necessary because Winamp could crash between the
					     // last call to GetWinampInstance() and this message send.

			DWORD MessageResult;
			int alive = SendMessageTimeout(WinampHWND,WM_COMMAND,STOP,0, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

			if (alive == 0)
				Trace::WriteLine("Winamp did not respond to the STOP message; the message has been discarded.  (12)");

			result = MessageResult;

			// changed:  result = SendMessage(WinampHWND,WM_COMMAND,STOP,0);

			if (diagnosticlevel >= 2)
				Trace::WriteLine("The response for the STOP message has been received.");

			if (diagnosticlevel >= 1)
				Trace::WriteLine("MediaPlayer state changed in WPStop() to STOPPED_STATE.");
		}
	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampPlugin_WPStop.");
	return;
}

/*
 * Class:     WinampPlugin
 * Method:    WPVolumeUp
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WinampPlugin_WPVolumeUp
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampPlugin_WPVolumeUp.");

	GetWinampInstance();

	if (WinampHWND)
	{
		LRESULT result;
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Sending RAISE_VOLUME message to Winamp and waiting for a response.");

		GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
				     // a message.  This is necessary because Winamp could crash between the
				     // last call to GetWinampInstance() and this message send.

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_COMMAND,RAISE_VOLUME,0, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the RAISE_VOLUME message; the message has been discarded.  (13)");

		result = MessageResult;

		// changed:  result = SendMessage(WinampHWND,WM_COMMAND,RAISE_VOLUME,0);

		if (diagnosticlevel >= 2)
			Trace::WriteLine("The response for the RAISE_VOLUME message has been received.");
	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampPlugin_WPVolumeUp.");
	return;
}


/*
 * Class:     WinampPlugin
 * Method:    WPVolumeDown
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_WinampPlugin_WPVolumeDown
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampPlugin_WPVolumeDown.");

	GetWinampInstance();

	if (WinampHWND)
	{
		LRESULT result;
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Sending LOWER_VOLUME message to Winamp and waiting for a response.");

		GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
				     // a message.  This is necessary because Winamp could crash between the
				     // last call to GetWinampInstance() and this message send.

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_COMMAND,LOWER_VOLUME,0, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the LOWER_VOLUME message; the message has been discarded.  (14)");

		result = MessageResult;

		// changed:  result = SendMessage(WinampHWND,WM_COMMAND,LOWER_VOLUME,0);

		if (diagnosticlevel >= 2)
			Trace::WriteLine("The response for the LOWER_VOLUME message has been received.");
	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampPlugin_WPVolumeDown.");
	return;
}

void CheckDiagnosticLevel()
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering CheckDiagnosticLevel.");

	char *thelevel = getenv("WINAMP_PLUGIN_DIAGNOSTIC_LEVEL");

	diagnosticlevel = 0;

	if (thelevel)
	{
		diagnosticlevel = atoi(thelevel);
	}

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting CheckDiagnosticLevel.");

	return;
}

/*
 * Class:     WinampPlugin
 * Method:    WPGetDiagnosticLevel
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_WinampPlugin_WPGetDiagnosticLevel
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampPlugin_WPGetDiagnosticLevel.");

	CheckDiagnosticLevel();

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampPlugin_WPGetDiagnosticLevel.");
	return diagnosticlevel;
}

/*
 * Class:     WinampPlugin
 * Method:    WPStatus
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_WinampPlugin_WPStatus
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampPlugin_WPStatus.");

	GetWinampInstance();

	if (state == EOS_STATE)
	{
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Playback ENDED (EOS_STATE) (REQUERY).");

		SendNudge();

		return EOS_STATE;
	}

	if (WinampHWND)
	{
		LRESULT result;
		LRESULT result1;
		LRESULT result2;

		// Is it time to check for EOS?
		if (GetTickCount() - LastLoad > TIME_ALLOWED_TO_LOAD)
		{
			if (diagnosticlevel >= 2)
				Trace::WriteLine("Loading grace period over; checking for EOS...");
			
			// are we supposed to be playing something?
			if (state == PLAY_STATE)
			{
				if (diagnosticlevel >= 2)
					Trace::WriteLine("A track is supposed to be playing.");

				// is it?
				if (diagnosticlevel >= 2)
					Trace::WriteLine("Sending CURRENT_STATUS message to Winamp and waiting for a response.");

				GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
						     // a message.  This is necessary because Winamp could crash between the
						     // last call to GetWinampInstance() and this message send.

				DWORD MessageResult;
				int alive = SendMessageTimeout(WinampHWND,WM_USER,0,CURRENT_STATUS, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

				if (alive == 0)
					Trace::WriteLine("Winamp did not respond to the CURRENT_STATUS message; the message has been discarded.  (15)");

				result = MessageResult;

				// changed:  result = SendMessage(WinampHWND,WM_USER,0,CURRENT_STATUS);

				if (diagnosticlevel >= 2)
					Trace::WriteLine("The response for the CURRENT_STATUS message has been received.");

				// it is NOT
				if (result != 1 && result != 3)
				{
					if (diagnosticlevel >= 2)
						Trace::WriteLine("...but there is no track playing.  Winamp must have stopped at EOS, so change state to EOS_STATE.");

					if (diagnosticlevel >= 1)
						Trace::WriteLine("MediaPlayer state changed in WPStatus() to EOS_STATE because Winamp has stopped playing the track.");
					state = EOS_STATE;

					if (diagnosticlevel >= 2)
						Trace::WriteLine("Playback ENDED (EOS_STATE).");

					SendNudge();
					return EOS_STATE;

				}

				// it is
				if (result == 1 || result == 3)
				{
					if (diagnosticlevel >= 2)
						Trace::WriteLine("...and there is...");
					if (diagnosticlevel >= 2)
						Trace::WriteLine("Sending CURRENT_TRACK_POSITION message to Winamp and waiting for a response.");

					GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
							     // a message.  This is necessary because Winamp could crash between the
							     // last call to GetWinampInstance() and this message send.

					DWORD MessageResult2;
					int alive2 = SendMessageTimeout(WinampHWND,WM_USER,0,CURRENT_TRACK_POSITION, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult2);

					if (alive2 == 0)
						Trace::WriteLine("Winamp did not respond to the CURRENT_TRACK_POSITION message; the message has been discarded.  (16)");

					result1 = MessageResult2;

					// changed:  result1 = SendMessage(WinampHWND,WM_USER,0,CURRENT_TRACK_POSITION);

					if (diagnosticlevel >= 2)
						Trace::WriteLine("The response for the CURRENT_TRACK_POSITION message has been received.");

					if (diagnosticlevel >= 2)
						Trace::WriteLine("Sending TOTAL_TRACK_DURATION message to Winamp and waiting for a response.");

					GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
							     // a message.  This is necessary because Winamp could crash between the
							     // last call to GetWinampInstance() and this message send.

					DWORD MessageResult3;
					int alive3 = SendMessageTimeout(WinampHWND,WM_USER,1,TOTAL_TRACK_DURATION, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult3);

					if (alive3 == 0)
						Trace::WriteLine("Winamp did not respond to the message; the message has been discarded.  (17)");

					result2 = MessageResult3;

					// changed:  result2 = SendMessage(WinampHWND,WM_USER,1,TOTAL_TRACK_DURATION);

					if (diagnosticlevel >= 2)
						Trace::WriteLine("The response for the TOTAL_TRACK_DURATION message has been received.");

					result2 *= 1000;
					
					// has the elapsed time exceeded the maximum time or reverted to 0?
					if (result1 == 0 || result1 >= result2)
					{
						if (diagnosticlevel >= 2)
							Trace::WriteLine("...but its elapsed time has reverted to 0 or exceeded the total track duration.");

						// it has
						state = EOS_STATE;

						if (diagnosticlevel >= 1)
							Trace::WriteLine("MediaPlayer state changed in WPStatus() to EOS_STATE because the elapsed track time is >= the total track time.");

						if (diagnosticlevel >= 2)
							Trace::WriteLine("Playback ENDED (ELAPSED_TIME).");
						SendNudge();
						return EOS_STATE;
					}
				}
			}
			else
			{
				if (diagnosticlevel >= 2)
					Trace::WriteLine("A track is NOT supposed to be playing.");
			}

			if (state == STOPPED_STATE)
			{
				if (diagnosticlevel >= 2)
					Trace::WriteLine("Playback STOPPED (STOPPED_STATE).");
				return STOPPED_STATE;
			} 

			// if winamp is playing, return a state of 2
			if (state == PLAY_STATE)
			{
				if (diagnosticlevel >= 2)
					Trace::WriteLine("Playback PLAYING (PLAY_STATE).");
				return PLAY_STATE;
			}

			// if winamp is paused, return a state of 3
			if (state == PAUSE_STATE)
			{
				if (diagnosticlevel >= 2)
					Trace::WriteLine("Playback PAUSED (PAUSE_STATE).");
				return PAUSE_STATE;
			}

			if (state == NO_STATE)
			{
				if (diagnosticlevel >= 2)
					Trace::WriteLine("Playback UNINITIALIZED (NO_STATE).");
				return NO_STATE;
			}
	
			if (state == LOADED_STATE)
			{
				if (diagnosticlevel >= 2)
					Trace::WriteLine("READY FOR PLAYBACK (LOADED_STATE).");
				return LOADED_STATE;
			}

		}
		else
		{
			if (diagnosticlevel >= 2)
				Trace::WriteLine("Track is still in its loading grace period; not checking for EOS.");
			return state;
		}

		if (diagnosticlevel >= 2)
			Trace::WriteLine("Exiting Java_WinampPlugin_WPStatus.");

		if (diagnosticlevel >= 2)
			Trace::WriteLine("WARNING:  MediaPlayer state unknown.  (forcing STOPPED_STATE)");
		return STOPPED_STATE;
	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampPlugin_WPStatus.");

	Trace::WriteLine("ERROR:  Winamp unavailable! (returning last known state)");
	return state;
}


/*
 * Class:     WinampPlugin
 * Method:    WPDuration
 * Signature: ()J
 */
JNIEXPORT jlong JNICALL Java_WinampPlugin_WPDuration
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampPlugin_WPDuration.");

	GetWinampInstance();

	if (WinampHWND)
	{
		LRESULT result;
		LRESULT result0;
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Sending TOTAL_TRACK_DURATION message to Winamp and waiting for a response.");

		GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
				     // a message.  This is necessary because Winamp could crash between the
				     // last call to GetWinampInstance() and this message send.

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_USER,1,TOTAL_TRACK_DURATION, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the TOTAL_TRACK_DURATION message; the message has been discarded.  (18)");

		result = MessageResult;

		// changed:  result = SendMessage(WinampHWND,WM_USER,1,TOTAL_TRACK_DURATION);

		if (diagnosticlevel >= 2)
			Trace::WriteLine("The response for the TOTAL_TRACK_DURATION message has been received.");

		if (diagnosticlevel >= 2)
			Trace::WriteLine("Sending CURRENT_TRACK_POSITION message to Winamp and waiting for a response.");

		GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
				     // a message.  This is necessary because Winamp could crash between the
				     // last call to GetWinampInstance() and this message send.

		DWORD MessageResult2;
		int alive2 = SendMessageTimeout(WinampHWND,WM_USER,0,CURRENT_TRACK_POSITION, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult2);

		if (alive2 == 0)
			Trace::WriteLine("Winamp did not respond to the CURRENT_TRACK_POSITION message; the message has been discarded.  (19)");

		result0 = MessageResult2;

		// changed:  result0 = SendMessage(WinampHWND,WM_USER,0,CURRENT_TRACK_POSITION);

		if (diagnosticlevel >= 2)
			Trace::WriteLine("The response for the CURRENT_TRACK_POSITION message has been received.");

		if (diagnosticlevel >= 2)
			Trace::WriteLine("Exiting Java_WinampPlugin_WPDuration.");

		// Never return a negative value.
		if (result >= 0)
			return (result * 1000);
		else
			return 0;

	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampPlugin_WPDuration.");
	return 0;
}

/*
 * Class:     WinampPlugin
 * Method:    WPElapsed
 * Signature: ()J
 */
JNIEXPORT jlong JNICALL Java_WinampPlugin_WPElapsed
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampPlugin_WPElapsed.");

	GetWinampInstance();

	if (WinampHWND)
	{
		LRESULT result;
		LRESULT result2;
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Sending CURRENT_TRACK_POSITION message to Winamp and waiting for a response.");

		GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
				     // a message.  This is necessary because Winamp could crash between the
				     // last call to GetWinampInstance() and this message send.

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_USER,0,CURRENT_TRACK_POSITION, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the CURRENT_TRACK_POSITION message; the message has been discarded.  (20)");

		result = MessageResult;

		// changed:  result = SendMessage(WinampHWND,WM_USER,0,CURRENT_TRACK_POSITION);

		if (diagnosticlevel >= 2)
			Trace::WriteLine("The response for the CURRENT_TRACK_POSITION message has been received.");

		if (diagnosticlevel >= 2)
			Trace::WriteLine("Sending TOTAL_TRACK_DURATION message to Winamp and waiting for a response.");

		GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
				     // a message.  This is necessary because Winamp could crash between the
				     // last call to GetWinampInstance() and this message send.

		DWORD MessageResult2;
		int alive2 = SendMessageTimeout(WinampHWND,WM_USER,1,TOTAL_TRACK_DURATION, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult2);

		if (alive2 == 0)
			Trace::WriteLine("Winamp did not respond to the TOTAL_TRACK_DURATION message; the message has been discarded.  (21)");

		result2 = MessageResult2;

		// changed:  result2 = SendMessage(WinampHWND,WM_USER,1,TOTAL_TRACK_DURATION);

		if (diagnosticlevel >= 2)
			Trace::WriteLine("The response for the TOTAL_TRACK_DURATION message has been received.");

		result2 *= 1000;

		if (diagnosticlevel >= 2)
			Trace::WriteLine("Exiting Java_WinampPlugin_WPElapsed.");

		// Never return a negative value.
		if (result >= 0)
			return result;
		else
			return 0;
	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampPlugin_WPElapsed.");
	return 0;
}

/*
 * Class:     WinampPlugin
 * Method:    WPSeek
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_WinampPlugin_WPSeek
  (JNIEnv *env, jobject jo, jlong SeekTime)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampPlugin_WPSeek.");

	GetWinampInstance();

	if (WinampHWND)
	{
		LRESULT result;
		LRESULT duration;
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Sending TOTAL_TRACK_DURATION message to Winamp and waiting for a response.");

		GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
				     // a message.  This is necessary because Winamp could crash between the
				     // last call to GetWinampInstance() and this message send.

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_USER,1,TOTAL_TRACK_DURATION, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the TOTAL_TRACK_DURATION message; the message has been discarded.  (22)");

		duration = MessageResult;

		// changed:  duration = SendMessage(WinampHWND,WM_USER,1,TOTAL_TRACK_DURATION);

		if (diagnosticlevel >= 2)
			Trace::WriteLine("The response for the TOTAL_TRACK_DURATION message has been received.");

		duration *= 1000;

		// Do not allow seeking past the end of the song.
		if (SeekTime > duration)
			SeekTime = duration;

		// Do not allow seeking to a negative position.
		if (SeekTime < 0)
			SeekTime = 0;

		if (diagnosticlevel >= 2)
			Trace::WriteLine("Sending SEEK message to Winamp and waiting for a response.");


		GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
				     // a message.  This is necessary because Winamp could crash between the
				     // last call to GetWinampInstance() and this message send.

		DWORD MessageResult2;
		int alive2 = SendMessageTimeout(WinampHWND,WM_USER,SeekTime,SEEK, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult2);

		if (alive2 == 0)
			Trace::WriteLine("Winamp did not respond to the SEEK message; the message has been discarded.  (23)");

		result = MessageResult2;

		// changed:  result = SendMessage(WinampHWND,WM_USER,SeekTime,SEEK);

		if (diagnosticlevel >= 2)
			Trace::WriteLine("The response for the SEEK message has been received.");

		if (diagnosticlevel >= 2)
			Trace::WriteLine("Exiting Java_WinampPlugin_WPSeek.");

		return SeekTime;
	}

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampPlugin_WPSeek.");

	return SeekTime;
}

/*
 * Class:     WinampPlugin
 * Method:    WPGetMuteStatus
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_WinampPlugin_WPGetMuteStatus
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampPlugin_WPGetMuteStatus.");

	GetWinampInstance();

	if (WinampHWND)
	{
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Exiting Java_WinampPlugin_WPGetMuteStatus.");

		if (muted == 1)
			return true;
		else
			return false;
	}

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampPlugin_WPGetMuteStatus.");

	return false;
}

/*
 * Class:     WinampPlugin
 * Method:    WPSetMuteStatus
 * Signature: (Z)V
 */
JNIEXPORT void JNICALL Java_WinampPlugin_WPSetMuteStatus
  (JNIEnv *env, jobject jo, jboolean MuteStatus)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampPlugin_WPSetMuteStatus.");

	GetWinampInstance();

	if (WinampHWND)
	{

		if (MuteStatus == JNI_TRUE)
		{
			muted = 1;
		}
		else
		{
			muted = 0;

			// Set the volume to the max.
			LRESULT result;
			if (diagnosticlevel >= 2)
				Trace::WriteLine("Sending SET_VOLUME_ABSOLUTE message to Winamp and waiting for a response.");

			GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
					     // a message.  This is necessary because Winamp could crash between the
					     // last call to GetWinampInstance() and this message send.

			DWORD MessageResult;
			int alive = SendMessageTimeout(WinampHWND,WM_USER,volume,SET_VOLUME_ABSOLUTE, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

			if (alive == 0)
				Trace::WriteLine("Winamp did not respond to the SET_VOLUME_ABSOLUTE message; the message has been discarded.  (24)");

			result = MessageResult;

			// changed:  result = SendMessage(WinampHWND,WM_USER,volume,SET_VOLUME_ABSOLUTE);

			if (diagnosticlevel >= 2)
				Trace::WriteLine("The response for the SET_VOLUME_ABSOLUTE message has been received.");
		}
	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampPlugin_WPSetMuteStatus.");
	return;
}

/*
 * Class:     WinampPlugin
 * Method:    WPGetVolume
 * Signature: ()F
 */
JNIEXPORT jfloat JNICALL Java_WinampPlugin_WPGetVolume
(JNIEnv *env, jobject jo)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampPlugin_WPGetVolume.");

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampPlugin_WPGetVolume.");

	if (muted == 1)
		return 0.0;
	else
		return ((float)volume / 255);
}

/*
 * Class:     WinampPlugin
 * Method:    WPSetVolume
 * Signature: (F)F
 */
JNIEXPORT jfloat JNICALL Java_WinampPlugin_WPSetVolume
  (JNIEnv *env, jobject jo, jfloat TheVolume)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering Java_WinampPlugin_WPSetVolume.");

	GetWinampInstance();

	if (WinampHWND)
	{
		LRESULT result;
		volume = (float)TheVolume * (float)255;

		if (diagnosticlevel >= 2)
			Trace::WriteLine("Sending SET_VOLUME_ABSOLUTE message to Winamp and waiting for a response.");


		GetWinampInstance(); // Added to force entry into mutex manager before ANY attempt to send
				     // a message.  This is necessary because Winamp could crash between the
				     // last call to GetWinampInstance() and this message send.

		DWORD MessageResult;
		int alive = SendMessageTimeout(WinampHWND,WM_USER,volume,SET_VOLUME_ABSOLUTE, SMTO_ABORTIFHUNG | SMTO_NORMAL, 15000, &MessageResult);

		if (alive == 0)
			Trace::WriteLine("Winamp did not respond to the SET_VOLUME_ABSOLUTE message; the message has been discarded.  (25)");

		result = MessageResult;

		// changed:  result = SendMessage(WinampHWND,WM_USER,volume,SET_VOLUME_ABSOLUTE);

		if (diagnosticlevel >= 2)
			Trace::WriteLine("The response for the SET_VOLUME_ABSOLUTE message has been received.");
	}
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting Java_WinampPlugin_WPSetVolume.");
	return TheVolume;
}

void SendNudge()
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering SendNudge().");

	if (GetTickCount() - LastNudge > TIME_BETWEEN_NUDGES)
	{
		LastNudge = GetTickCount();
	
		HWND SageTVHWND;	
		SageTVHWND = FindWindow("SageApp", "SageWin");
		if (SageTVHWND)
		{
			if (diagnosticlevel >= 1)
				Trace::WriteLine("Sending a nudge to SageTV to make it notice that the EOS_STATE is set.");
			PostMessage(SageTVHWND, WM_USER + 396, 0, 0);
		}
		else
		{
			Trace::WriteLine("An attempt to send a nudge failed because the SageTV window could not be found.");
		}
	}
	else
	{
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Not sending nudge because one was sent recently.");
	}

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting SendNudge().");
	return;
}

/*

// Uncomment the following code to support international characters.

void JNU_ThrowByName(JNIEnv *env, const char *name, const char *msg)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering JNU_ThrowByName().");

	jclass cls;

	cls = env->FindClass(name);

	// if cls is NULL, an exception has already been thrown

	if (cls != NULL) {
		env->ThrowNew(cls, msg);
	}

	// free the local ref
	env->DeleteLocalRef(cls);

	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting JNU_ThrowByName().");
}

char *JNU_GetStringNativeChars(JNIEnv *env, jstring jstr)
{
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Entering JNU_GetStringNativeChars().");

	jbyteArray bytes = 0;
	jthrowable exc;
	char *result = 0;
	jclass jcls_str;
	jmethodID MID_String_getBytes;

	if (env->EnsureLocalCapacity(2) < 0) {
		if (diagnosticlevel >= 2)
			Trace::WriteLine("Exiting JNU_GetStringNativeChars() due to out-of-memory error (1).");
		return 0; // out of memory error
	}

	jcls_str = env->FindClass("java/lang/String");

	MID_String_getBytes = env->GetMethodID(jcls_str, "getBytes", "()[B");

	bytes = (jbyteArray)env->CallObjectMethod(jstr, MID_String_getBytes);
	exc = env->ExceptionOccurred();

	if (!exc) {
		jint len = env->GetArrayLength(bytes);
		result = (char *)malloc(len + 1);

		if (result == 0) {
			JNU_ThrowByName(env, "java/lang/OutOfMemoryError", 0);
			env->DeleteLocalRef(bytes);
			if (diagnosticlevel >= 2)
				Trace::WriteLine("Exiting JNU_GetStringNativeChars() due to out-of-memory error (2).");
			return 0;
		}

		env->GetByteArrayRegion(bytes, 0, len, (jbyte *)result);

		result[len] = 0; // NULL-terminate
	} else {
		env->DeleteLocalRef(exc);
	}

	env->DeleteLocalRef(bytes);
	if (diagnosticlevel >= 2)
		Trace::WriteLine("Exiting JNU_GetStringNativeChars().");
	return (char*)result;
}

*/

