public class WinampPlugin implements sage.MediaPlayer
{
	// Media Player States
	int NO_STATE = 0;
	int LOADED_STATE = 1;
	int PLAY_STATE = 2;
	int PAUSE_STATE = 3;
	int STOPPED_STATE = 4;
	int EOS_STATE = 5;

	// Closed Caption States
	int CC_DISABLED = 0;
	int CC_ENABLED_CAPTION1 = 1;
	int CC_ENABLED_CAPTION2 = 2;
	int CC_ENABLED_TEXT1 = 3;
	int CC_ENABLED_TEXT2 = 4;

	// Capabilities (as of SageTV 4.x, these do nothing)
	int FRAME_STEP_FORWARD_CAP = 0x1;
	int FRAME_STEP_BACKWARD_CAP = 0x2;
	int PAUSE_CAP = 0x4;
	int PLAYRATE_FAST_CAP = 0x8;
	int PLAYRATE_SLOW_CAP = 0x10;
	int PLAY_REV_CAP = 0x20;
	int PLAYRATE_FAST_REV_CAP = 0x40;
	int PLAYRATE_SLOW_REV_CAP = 0x80;
	int SEEK_CAP = 0x100;
	int TIMESHIFT_CAP = 0x200;
	int LIVE_CAP = 0x400;

	java.io.File CurrentFile = null;
	java.awt.Dimension VideoDimensions = new java.awt.Dimension(0,0);
	int DiagnosticLevel = 0;

	static
	{
		System.out.println("WINAMP_PLUGIN:  The WinampPlugin class is attempting to load the WinampJNI.dll library.");

		try
		{
			System.loadLibrary("WinampJNI");
		}
		catch (SecurityException s)
		{
			System.out.println("WINAMP_PLUGIN:  The WinampJNI.dll library could not be loaded due to a security exception.");
		}
		catch (UnsatisfiedLinkError l)
		{
			System.out.println("WINAMP_PLUGIN:  The WinampJNI.dll library could not be loaded due to an unsatisfied link error (ie: the file was not found).");
		}

		System.out.println("WINAMP_PLUGIN:  The WinampPlugin class has loaded WinampJNI.dll library without exception.");
	}

	// JNI Functions; these are the functions in the JNI .dll file
	// that will actually do the work.
	public native void WPPlayFile(String TheFile);
	public native void WPStop();
	public native void WPPlay();
	public native void WPPause();
	public native int WPStatus();
	public native long WPDuration();
	public native long WPElapsed();
	public native long WPSeek(long SeekPoint);
	public native boolean WPGetMuteStatus();
	public native void WPSetMuteStatus(boolean MuteStatus);
	public native float WPGetVolume();
	public native float WPSetVolume(float TheVolume);
	public native void WPStartup();
	public native void WPShutdown();
	public native int WPGetDiagnosticLevel();

	public WinampPlugin()
	{
		System.out.println("WINAMP_PLUGIN:  A new instance of the WinampPlugin class is being instantiated.");

		System.out.println("WINAMP_PLUGIN:  Initializing the WinampJNI.dll library (startup).");

		WPStartup();

		System.out.println("WINAMP_PLUGIN:  The WinampJNI.dll library has been initialized.");

		System.out.println("WINAMP_PLUGIN:  Determining the diagnostic level.");

		DiagnosticLevel = WPGetDiagnosticLevel();

		System.out.println("WINAMP_PLUGIN:  The diagnostic level has been determined.");

		if (DiagnosticLevel == 0)
			System.out.println("WINAMP_PLUGIN:  Diagnostic Level = 0 (Minimal Tracing)");

		if (DiagnosticLevel == 1)
			System.out.println("WINAMP_PLUGIN:  Diagnostic Level = 1 (Basic Tracing)");

		if (DiagnosticLevel == 2)
			System.out.println("WINAMP_PLUGIN:  Diagnostic Level = 2 (Verbose Tracing)");
	}

	public void load(byte majorTypeHint, byte minorTypeHint, String encodingHint, java.io.File file, String hostname, boolean timeshifted, long bufferSize) throws sage.PlaybackException
	{
		/*
		 * When called the MediaPlayer should load the specified file.
		 *
		 * @param majorTypeHint Indicates the type of content (major):
		 *
	         * 	UNKNOWN = 0;
		 * 	VIDEO = 1;
		 * 	AUDIO = 2;
		 * 	DVD = 4;
		 * 	CD = 5;
		 *
		 * @param minorTypeHint Indicates the type of content (minor):
		 *
	         * 	UNKNOWN = 30;
		 * 	MPEG2_PS = 31;
		 * 	MPEG2_TS = 32;
		 * 	MP3 = 33;
		 * 	MPEG1_PS = 36;
		 * 	WAV = 38;
		 * 	AVI = 39;
	         * 	MPEG4->MPEG2 = 40;
	         * 	DIVX->MPEG2 = 41;
		 *
		 * @param encodingHint A descriptive string for the format the content was recorded in, do not rely on this.
		 *
		 * @param file The file to play.
		 *
		 * @param hostname null for local files, for remote files it's the IP hostname/address that the file is located on.
		 *
		 * @param timeshifted true if the file is currently being recorded
		 *
		 * @param bufferSize The size of the circular buffer used to record to the file being played back.
		 *
		 */	

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In load() now.");

		if (DiagnosticLevel>=1)
			System.out.println("WINAMP_PLUGIN:  The name of the file to play is '" + file.toString() + "'.  The hostname where the file is located is '" + hostname + "'.");

		try
		{
			CurrentFile = file;
			WPPlayFile(file.toString());
		}
		catch (Exception ex)
		{
			System.out.println("WINAMP_PLUGIN:  An EXCEPTION has been detected.");
		}

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  Exiting load() now." + "  " + "This function has no return value (void).");
		return;
	}

	public void fastLoad(byte majorTypeHint, byte minorTypeHint, String encodingHint, java.io.File file, String hostname, boolean timeshifted, long bufferSize, boolean waitUntilDone) throws sage.PlaybackException
	{
		// This MediaPlayer does not understand the concept of "fast loading".  
		// The canFastLoad() method will always return false, so this method does not need to be implemented.

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In fastLoad() now.");

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  Exiting fastLoad() now." + "  " + "This function has no return value (void).");

		return;
	}

	public boolean canFastLoad(byte majorTypeHint, byte minorTypeHint, String encodingHint, java.io.File file)
	{
		// This MediaPlayer does not understand the concept of "fastload", so return false.

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In canFastLoad() now.");

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  Exiting canFastLoad() now." + "  " + "return value = 'false'.");
		return false;
	}

	public void free()
	{
		// This MediaPlayer uses no resources; if it did, this would be the place to release them.

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In free() now.");

		if (DiagnosticLevel>=1)
			System.out.println("WINAMP_PLUGIN:  Media playback complete; SageTV has instructed the media player to release all resources.");

		try
		{
			// Halt playback.
			stop();

			// Stop referncing the file object that SageTV passed in so that it can be garbage collected.
			CurrentFile = null;
		}
		catch (Exception ex)
		{
			System.out.println("WINAMP_PLUGIN:  An EXCEPTION has been detected.");
		}

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  Exiting free() now." + "  " + "This function has no return value (void).");
		return;
	}

	public void inactiveFile()
	{
		// This plugin doesn't care about timeshifting.  If it did, this method would be called when the file
		// currently being played back became inactive (ie: was no longer being written to).

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In inactiveFile() now.");

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  Exiting inactiveFile() now." + "  " + "This function has no return value (void).");
		return;
	}
    
	public void stop()
	{
		// Stop playback.
		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In stop() now.");
		try
		{
			WPStop();
		}
		catch (Exception ex)
		{
			System.out.println("WINAMP_PLUGIN:  An EXCEPTION has been detected.");
		}

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  Exiting stop() now." + "  " + "This function has no return value (void).");
		return;
	}

	public java.io.File getFile()
   	{
		// Return a java.io.file object representing the file that the MediaPlayer is currently playing.
		// I simply store the object passed into load() and return it here.  I'm not sure if this is a good
		// or bad idea, but it seems like a reasonable thing to do for a MediaPlayer that isn't capable of
		// spanning files.
		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In getFile() now.");
		try
		{
			if (DiagnosticLevel>=2)
				System.out.println("WINAMP_PLUGIN:  Exiting getFile() now." + "  " + "return value='" + CurrentFile.toString() + "'.");
			return CurrentFile;
		}
		catch (Exception ex)
		{
			System.out.println("WINAMP_PLUGIN:  An EXCEPTION has been detected.");

			if (DiagnosticLevel>=2)
				System.out.println("WINAMP_PLUGIN:  Exiting getFile() now." + "  " + "return value = (default) null.");

			return null;
		}
	}

	public int getState()
	{
		// Return the current state of the MediaPlayer (playing, paused, stopped, etc.).
		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In getState() now.");

		try
		{
			int TheState;
			TheState = WPStatus();

			if (DiagnosticLevel>=2)
				System.out.println("WINAMP_PLUGIN:  Exiting getState() now." + "  " + "return value = '" + TheState + "'.");
			return TheState;
		}
		catch (Exception ex)
		{
			System.out.println("WINAMP_PLUGIN:  An EXCEPTION has been detected.");

			if (DiagnosticLevel>=2)
				System.out.println("WINAMP_PLUGIN:  Exiting getState() now due to exception." + "  " + "return value = (default) NO_STATE (0).");

			return NO_STATE;
		}
	}

	public long getDurationMillis()
	{
		// Return the duration of the item that the MediaPlayer is playing, in milliseconds.
		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In getDurationMillis() now.");

		try
		{
			long TheDuration;

			TheDuration = WPDuration();

			/*
			int FailCount = 0;
			int FailMaximum = 30000;

			while (FailCount < FailMaximum && TheDuration == 0)
			{
				System.out.println("WINAMP_PLUGIN:  Attempting to get duration.");
				TheDuration = WPDuration();

				if (FailCount >= FailMaximum - 1)
				{
					System.out.println("WINAMP_PLUGIN:  Winamp has failed to report the file duration after " + FailMaximum + " attempts.  forcing return value of '30000'.");
					return 300000;
				}
				FailCount++;
			}
			System.out.println("WINAMP_PLUGIN:  Winamp returned the duration after " + FailCount + " attempts.");
			*/

			if (DiagnosticLevel>=2)
				System.out.println("WINAMP_PLUGIN:  Exiting getDurationMillis() now." + "  " + "return value = '" + TheDuration + "'.");
			return TheDuration;
		}
		catch (Exception ex)
		{
			System.out.println("WINAMP_PLUGIN:  An EXCEPTION has been detected.");

			if (DiagnosticLevel>=2)
				System.out.println("WINAMP_PLUGIN:  Exiting getDurationMillis() now due to exception." + "  " + "return value = (default) 0.");

			return 0;
		}
	}

	public long getMediaTimeMillis()
	{
		// Return the elapsed time of the item that the MediaPlayer is playing, in milliseconds.
		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In getMediaTimeMillis() now.");

		try
		{
			long TheElapsedTime;

			TheElapsedTime = WPElapsed();

			if (DiagnosticLevel>=2)
				System.out.println("WINAMP_PLUGIN:  Exiting getMediaTimeMillis() now." + "  " + "return value = '" + TheElapsedTime + "'.");
			return TheElapsedTime;
		}
		catch (Exception ex)
		{
			System.out.println("WINAMP_PLUGIN:  An EXCEPTION has been detected.");

			if (DiagnosticLevel>=2)
				System.out.println("WINAMP_PLUGIN:  Exiting getMediaTimeMillis() now due to exception." + "  " + "return value = (default) 0.");

			return 0;
		}
	}

	public float getPlaybackRate()
	{
		// Return the playback rate for the item that the MediaPlayer is playing.  Since this
		// MediaPlayer plays only at one speed, return 1.0.
		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In getPlaybackRate() now.");

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  Exiting getPlaybackRate() now." + "  " + "return value = '1.0'.");
		return (float)1.0;
	}

	public boolean pause()
	{
		// Pause playback.
		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In pause() now.");

		try
		{
			WPPause();
		}
		catch (Exception ex)
		{
			System.out.println("WINAMP_PLUGIN:  An EXCEPTION has been detected.");
		}

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  Exiting pause() now." + "  " + "return value = 'true'.");
		return true;
	}


	public boolean play()
	{
		// Start playback.

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In play() now.");

		try
		{
			WPPlay();
		}
		catch (Exception ex)
		{
			System.out.println("WINAMP_PLUGIN:  An EXCEPTION has been detected.");
		}

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  Exiting play() now." + "  " + "return value = 'true'.");

		return true;
	}
	
	public long seek(long seekTimeMillis) throws sage.PlaybackException
	{
		// Seek to a specific point in the item that the MediaPlayer is playing.
		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In seek() now.");

		try
		{
			long TheNewPosition;

			TheNewPosition = WPSeek(seekTimeMillis);

			if (DiagnosticLevel>=2)
				System.out.println("WINAMP_PLUGIN:  Exiting seek() now." + "  " + "return value = '" + TheNewPosition + "'.");

			return TheNewPosition;
		}
		catch (Exception ex)
		{
			System.out.println("WINAMP_PLUGIN:  An EXCEPTION has been detected.");

			if (DiagnosticLevel>=2)
				System.out.println("WINAMP_PLUGIN:  Exiting seek() now due to exception." + "  " + "return value = (default) 0.");

			return 0;
		}
	}

	public float setPlaybackRate(float newRate)
	{
		// ummm.... no.  This MediaPlayer does not support variable rate playback.  Returning 1.0
		// tells SageTV that the new rate is 1.0 -- which probably comes as a surprise to SageTV since 
		// it wanted something else, but oh well.  We don't always get what we want.

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In setPlaybackRate() now.");

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  Exiting setPlaybackRate() now." + "  " + "return value = '1.0'.");
		return (float)1.0;
	}

	public boolean frameStep(int amount)
	{
		// This plugin only deals with audio data.  Just return true.
		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In frameStep() now.");

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  Exiting frameStep() now." + "  " + "return value = 'true'.");
		return true;
	}

	public boolean getMute()
	{
		// Return whether or not the MediaPlayer is muted.
		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In getMute() now.");

		try
		{
			boolean TheMuteStatus;

			TheMuteStatus = WPGetMuteStatus();

			if (DiagnosticLevel>=2)
				System.out.println("WINAMP_PLUGIN:  Exiting getMute() now." + "  " + "return value = '" + TheMuteStatus + "'.");
			return TheMuteStatus;
		}
		catch (Exception ex)
		{
			System.out.println("WINAMP_PLUGIN:  An EXCEPTION has been detected.");

			if (DiagnosticLevel>=2)
				System.out.println("WINAMP_PLUGIN:  Exiting getMute() now due to exception." + "  " + "return value = (default) false.");

			return false;
		}
	}

	public void setMute(boolean x)
	{
		// Set the mute status to whatever SageTV wants it set to (true or false).
		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In setMute() now.");

		try
		{
			WPSetMuteStatus(x);
		}
		catch (Exception ex)
		{
			System.out.println("WINAMP_PLUGIN:  An EXCEPTION has been detected.");
		}

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  Exiting setMute() now." + "  " + "This function has no return value (void).");
		return;
	}

	public float getVolume()
	{
		// Return the volume level of the MediaPlayer.  This is a number between 0.0 and 1.0.  If your source
		// treats volume as a percent, multiply it by .001.  If your source treats volume as a number between 0
		// and X (like Winamp) then divide the volume by X and return the result. 
		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In getVolume() now.");

		try	
		{
			float TheVolume;

			TheVolume = WPGetVolume();

			if (DiagnosticLevel>=2)
				System.out.println("WINAMP_PLUGIN:  Exiting getVolume() now." + "  " + "return value = '" + TheVolume + "'.");
			return TheVolume;
		}
		catch (Exception ex)
		{
			System.out.println("WINAMP_PLUGIN:  An EXCEPTION has been detected.");
			if (DiagnosticLevel>=2)
				System.out.println("WINAMP_PLUGIN:  Exiting getVolume() now due to exception." + "  " + "return value = (default) 1.0.");
			return (float)1.0;
		}
	}

	public float setVolume(float f)
	{
		// Set the volume for the MediaPlayer.  Return the actual volume level that got set.  
		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In setVolume() now.");

		try
		{
			float TheNewVolume;
	
			TheNewVolume = WPSetVolume(f);

			if (DiagnosticLevel>=2)
				System.out.println("WINAMP_PLUGIN:  Exiting setVolume() now." + "  return value = '" + TheNewVolume + "'.");
			return TheNewVolume;
		}
		catch (Exception ex)
		{
			System.out.println("WINAMP_PLUGIN:  An EXCEPTION has been detected.");
			if (DiagnosticLevel>=2)
				System.out.println("WINAMP_PLUGIN:  Exiting setVolume() now due to exception.  return value = (default) 1.0.");
			return (float)1.0;
		}
	}
    
	public int getPlaybackCaps()
	{
		// Unused by SageTV at this time.  Per documentaiton, return 0.
		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In getPlaybackCaps() now.");

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  Exiting getPlaybackCaps() now." + "  " + "return value = '0'.");
		return 0;
	}
		
	public java.awt.Color getColorKey()
	{

		// This MediaPlayer plugin does not have a visual component.  Winamp will handle that.  If it did, 
		// this function would return the color to use as the color key.
		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In getColorKey() now.");

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  Exiting getColorKey() now." + "  " + "return value = 'null'.");
		return null;
	}
	
	public void setVideoRectangles(java.awt.Rectangle videoSrcRect, java.awt.Rectangle videoDestRect, boolean hideCursor)
	{
		// This MediaPlayer plugin does not have a visual component.
		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In setVideoRectangles() now.");

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  Exiting setVideoRectangles() now." + "  " + "This function has no return value (void).");
		return;
	}

	public java.awt.Dimension getVideoDimensions()
	{
		// Return the size of the video frame for the item being played back.  Since this MediaPlayer has no visual
		// component, return 0,0.
		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In getVideoDimensions() now.");

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  Exiting getVideoDimensions() now." + "  " + "return value = '0,0'.");
		return VideoDimensions;
	}
	
	public boolean setClosedCaptioningState(int ccState)
	{
		// This MediaPlayer plugin does not support closed captioning, so pretend like all is well.
		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In setClosedCaptioningState() now.");

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  Exiting setClosedCaptioningState() now." + "  " + "return value = 'true'.");

		return true;
	}

	public int getClosedCaptioningState()
	{
		// This MediaPlayer plugin does not support closed captioning, so return CC_DISABLED.
		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In getClosedCaptioningState() now.");

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  Exiting getClosedCaptioningState() now." + "  " + "return value = 'CC_DISABLED (0)'.");

		return CC_DISABLED;
	}

	public int getTransparency()
	{
		// This is a method in java.awt.transparency that must be overridden that indicates what type of 
		// transparency that the MediaPlayer plugin supports.  It can only be OPAQUE or BITMASK as of
		// SageTV 4.x.
		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  In getTransparency() now.");

		if (DiagnosticLevel>=2)
			System.out.println("WINAMP_PLUGIN:  Exiting getTransparency() now." + "  " + "return value = '" + OPAQUE + "'.");
		return OPAQUE;
	}
}

